import numpy as np
from scipy.spatial.distance import cdist
from joblib import Parallel, delayed

#01 Define spatial weight matrix
class w_matrix:
    @staticmethod
    def _compute_distance_matrix(df, xcoord, ycoord, metric='euclidean', **kwargs):
        points = df[[xcoord, ycoord]].values
        dismatrix = cdist(points, points, metric=metric, **kwargs)
        return dismatrix

    @staticmethod
    def _compute_weights_row(i, dismatrix, bandwidths, kernel_type):
        if kernel_type == 'Binary':
            return (dismatrix[i] <= bandwidths[i]).astype(float)
        elif kernel_type == 'Gaussian':
            return np.exp(-dismatrix[i]**2 / (2 * bandwidths[i]**2))
        elif kernel_type == 'GaussianBinary':
            gaussian_weights = np.exp(-dismatrix[i]**2 / (2 * bandwidths[i]**2))
            binary_weights = (dismatrix[i] <= bandwidths[i]).astype(float)
            return gaussian_weights * binary_weights
    @staticmethod
    def spatial_weight(df, xcoord, ycoord, fix=False, bandwidth=10, kernel_type='Binary', n_jobs=-1):
        dismatrix = w_matrix._compute_distance_matrix(df, xcoord, ycoord)
        bandwidths = np.zeros(dismatrix.shape[0])
        if fix:
            bandwidths.fill(bandwidth)
        else:
            for i in range(dismatrix.shape[0]):
                sorted_distances = np.sort(dismatrix[i])
                bandwidths[i] = sorted_distances[min(bandwidth, len(sorted_distances)-1)]
        weights = Parallel(n_jobs=n_jobs)(
            delayed(w_matrix._compute_weights_row)(i, dismatrix, bandwidths, kernel_type) for i in range(dismatrix.shape[0])
        )
        return np.array(weights)**0.5
    def from_libpysal(w):
        W_matrix, ids = w.full()
        W_symmetric = np.maximum(W_matrix, W_matrix.T)
        np.fill_diagonal(W_symmetric, 1)
        return W_symmetric**0.5


import numpy as np
import lime.lime_tabular
import shap

# 02 Define preprocess data function and explainer
class preprocess:
    def _preprocess_data(df, feature_names, target_name, point_index, weights):
        weight = weights[point_index]
        X_weighted = df[feature_names].multiply(weight, axis=0)
        y_weighted = df[target_name] * weight
        non_zero_indices = np.nonzero(weight)[0]
        selected_points = non_zero_indices[non_zero_indices != point_index]
        X_train = X_weighted.iloc[selected_points].values
        y_train = y_weighted.iloc[selected_points]
        current_point = X_weighted.iloc[point_index].values.reshape(1, -1)
        return X_train, y_train, current_point

    def _calculate_lime_importances(model, X_train, X_test, feature_names, target_name):
        explainer = lime.lime_tabular.LimeTabularExplainer(
            training_data=X_train,
            feature_names=feature_names,
            class_names=[target_name],
            mode='regression',
            discretize_continuous=False,
        )
        exp = explainer.explain_instance(data_row=X_test[0], predict_fn=model.predict)
        lime_importances_dict = dict(exp.as_list())
        importances = np.array([lime_importances_dict.get(feature, 0) for feature in feature_names])
        return importances

    def _calculate_shap_importances(model, X_train, X_test):
        explainer = shap.Explainer(model, X_train)
        shap_values = explainer(X_test)
        return np.abs(shap_values.values).mean(axis=0)


import pandas as pd
import numpy as np
from tqdm import tqdm
from joblib import Parallel, delayed

# 03 Quick train and predict without keep models
class fasttrain:
    def _explain_point(df, feature_names, target_name, point_index, weights, model_class, explainer_names):
        X_train, y_train, X_test =preprocess._preprocess_data(df, feature_names, target_name, point_index, weights)
        model = model_class()  # Assuming model_class is defined elsewhere
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)[0]
        importances = {}
        if 'LIME' in explainer_names:
            importances['LIME'] = preprocess._calculate_lime_importances(model, X_train, X_test, feature_names, target_name)
        if 'SHAP' in explainer_names:
            importances['SHAP'] = preprocess._calculate_shap_importances(model, X_train, X_test)            
        if 'Importance' in explainer_names and hasattr(model, 'feature_importances_'):
            importances['Importance'] = model.feature_importances_
        importance_combine = np.concatenate([value.reshape(-1) for value in importances.values()])
        return y_pred, importance_combine

    def _predict_point(df, feature_names, target_name, point_index, weights, model_class):
        X_train, y_train, X_test = preprocess._preprocess_data(df, feature_names, target_name, point_index, weights)
        model = model_class()  # Assuming model_class is defined elsewhere
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)[0]
        return y_pred

    def explain_all(df, feature_names, target_name, weights, model_class, explainer_names):
        #pretest
        X_train1, y_train1, _ =preprocess._preprocess_data(df, feature_names, target_name, 1, weights)
        model = model_class()  # Assuming model_class is defined elsewhere
        model.fit(X_train, y_train)
        model_supports_importance = hasattr(model, 'feature_importances_')
        explainerlist= explainer_names.copy()
        if 'Importance' in explainer_names and not model_supports_importance:
            explainerlist.remove('Importance')
        predictions = np.zeros(len(df))
        feature_importances_matrix = np.zeros((len(df), len(feature_names) * len(explainerlist)))
        results = Parallel(n_jobs=-1)(
            delayed(fasttrain._explain_point)(df, feature_names, target_name, i, weights, model_class, explainerlist) for i in tqdm(range(len(df)), desc="Computing Prediction")
        )
        for i, (y_pred, importances) in enumerate(results):
            predictions[i] = y_pred
            feature_importances_matrix[i, :] = importances
        columns = [var + '_' + explainer for explainer in explainerlist for var in feature_names]
        importances_df = pd.DataFrame(feature_importances_matrix, columns=columns)
        importances_df['predy'] = predictions
        df_new = pd.concat([df.reset_index(drop=True), importances_df.reset_index(drop=True)], axis=1)
        return df_new

    def pred_all(df, feature_names, target_name, weights, model_class):
        predictions = np.zeros(len(df))
        results = Parallel(n_jobs=-1)(
            delayed(fasttrain._predict_point)(df, feature_names, target_name, i, weights, model_class) for i in tqdm(range(len(df)), desc="Computing Prediction")
        )
        for i, y_pred in enumerate(results):
            predictions[i] = y_pred
        df['predy'] = predictions
        return  df

    def _generate_samples(df,feature_names, num_samples=50):
        sampled_df = pd.DataFrame()
        sample_var = pd.DataFrame()

        for col in feature_names:
            min_val, max_val = df[col].min(), df[col].max()
            samples = np.linspace(min_val, max_val, num_samples)

            for val in samples:
                subset = df[(df[col] >= val) & (df[col] < val + (max_val - min_val) / num_samples)]
                if not subset.empty:
                    sampled_df = pd.concat([sampled_df, subset.sample(n=1)])
            sample_var[col] = samples

        sampled_df.drop_duplicates(inplace=True)
        sampled_df=sampled_df[feature_names]
        return sampled_df, sample_var

    def _point_partial_dependent(df, point_index, sample_var, model_class,  feature_names, target_name, weights):
        X_train, y_train, X_test1 = preprocess._preprocess_data(df, feature_names, target_name, point_index, weights)
        tmodel = model_class()
        tmodel.fit(X_train, y_train)

        y_pred_matrix = np.zeros((sample_var.shape[0], len(feature_names) + 1))
        for i, feature_name in enumerate(feature_names):
            for j, val in enumerate(sample_var[feature_name]):
                X_test = X_test1.copy()
                X_test[0, i] = val
                y_pred = tmodel.predict(X_test)[0]
                y_pred_matrix[j, i] = y_pred
        y_pred_matrix[:, -1] = range(sample_var.shape[0])

        return y_pred_matrix

    def all_partial_dependent(df,  model_class,  feature_names, target_name, weights,num_samples=50):
        sampled_df, sample_var=fasttrain._generate_samples(df,feature_names, num_samples)
        print(f'Sample size: {num_samples}. {len(sampled_df)} used for Partial Dependence Estimation')
        results = Parallel(n_jobs=4)(delayed(fasttrain._point_partial_dependent)(df, i, sample_var, model_class,  feature_names, target_name, weights) for i in tqdm(sampled_df.index, desc="Computing importances"))
        merged_array = np.vstack(results)
        feature_id=feature_names.copy()
        feature_id.append('ID')
        resultsdf=pd.DataFrame(merged_array, columns=feature_id)
        agg_dict = {col: 'mean' for col in feature_id}
        results_mean = resultsdf.groupby('ID').agg(agg_dict)
        results_mean.drop(columns=['ID'],inplace=True)
        results_mean.columns = [f"{col}_estimate" for col in results_mean.columns]
        sample_var.columns= [f"{col}_sample" for col in sample_var.columns]
        df_pd= pd.concat([sample_var.reset_index(drop=True), results_mean.reset_index(drop=True)], axis=1)
        return  df_pd


from setuptools import setup, find_packages

setup(
    name='XGeoML',
    version='0.1.1',
    packages=find_packages(),
    description='A ensemble framework for explainable geospatial machine Learning models',
    author='Lingbo Liu',
    author_email='lingbo.liu@whu.edu.cn',
    license="MIT",
    install_requires=[
        'scikit-learn',
        'numpy',
        'scipy',
        'joblib',
        'pandas',
        'lime',
        'shap',
        'tqdm', 
    ],
)
